from types import MethodType
from warnings import warn

from scipy.stats import norm

import numpy as np
import pandas as pd

from . import common_args
from ..util import read_param_file, ResultDict, extract_group_names, _check_groups

from multiprocessing import Pool, cpu_count
from functools import partial
from itertools import combinations, zip_longest


CONST_RESULT_MSG = (
    "Constant values encountered, indicating model evaluations "
    "(or subset of evaluations) produced identical values."
)


def analyze(
    problem,
    Y,
    calc_second_order=True,
    num_resamples=100,
    conf_level=0.95,
    print_to_console=False,
    parallel=False,
    n_processors=None,
    keep_resamples=False,
    seed=None,
):
    """Perform Sobol Analysis on model outputs.

    Returns a dictionary with keys 'S1', 'S1_conf', 'ST', and 'ST_conf', where
    each entry is a list of size D (the number of parameters) containing the
    indices in the same order as the parameter file.  If calc_second_order is
    True, the dictionary also contains keys 'S2' and 'S2_conf'.


    Notes
    -----
    Compatible with:
        `saltelli` : :func:`SALib.sample.saltelli.sample`
        `sobol` : :func:`SALib.sample.sobol.sample`


    Examples
    --------
        >>> X = saltelli.sample(problem, 512)
        >>> Y = Ishigami.evaluate(X)
        >>> Si = sobol.analyze(problem, Y, print_to_console=True)


    Parameters
    ----------
    problem : dict
        The problem definition
    Y : numpy.array
        A NumPy array containing the model outputs
    calc_second_order : bool
        Calculate second-order sensitivities (default True)
    num_resamples : int
        The number of resamples (default 100)
    conf_level : float
        The confidence interval level (default 0.95)
    print_to_console : bool
        Print results directly to console (default False)
    parallel : bool
        Perform analysis in parallel if True
    n_processors : int
        Number of parallel processes (only used if parallel is True)
    keep_resamples : bool
        Whether or not to store intermediate resampling results (default False)
    seed : int
        Seed to generate a random number

    References
    ----------
    1. Sobol, I. M. (2001).
       Global sensitivity indices for nonlinear
       mathematical models and their Monte Carlo estimates.
       Mathematics and Computers in Simulation, 55(1-3):271-280,
       doi:10.1016/S0378-4754(00)00270-6.

    2. Saltelli, A. (2002).
       Making best use of model evaluations to compute sensitivity indices.
       Computer Physics Communications, 145(2):280-297
       doi:10.1016/S0010-4655(02)00280-1.

    3. Saltelli, A., P. Annoni, I. Azzini, F. Campolongo, M. Ratto, and
       S. Tarantola (2010).
       Variance based sensitivity analysis of model output. Design and
       estimator for the total sensitivity index.
       Computer Physics Communications, 181(2):259-270,
       doi:10.1016/j.cpc.2009.09.018.
    """
    if seed:
        # Set seed to ensure CIs are the same
        rng = np.random.default_rng(seed).integers
    else:
        rng = np.random.randint

    # determining if groups are defined and adjusting the number
    # of rows in the cross-sampled matrix accordingly
    groups = _check_groups(problem)
    if not groups:
        D = problem["num_vars"]
    else:
        _, D = extract_group_names(groups)

    if calc_second_order and Y.size % (2 * D + 2) == 0:
        N = int(Y.size / (2 * D + 2))
    elif not calc_second_order and Y.size % (D + 2) == 0:
        N = int(Y.size / (D + 2))
    else:
        raise RuntimeError(
            """
        Incorrect number of samples in model output file.
        Confirm that calc_second_order matches option used during sampling."""
        )

    if not 0 < conf_level < 1:
        raise RuntimeError("Confidence level must be between 0-1.")

    # normalize the model output
    Y = (Y - Y.mean()) / Y.std()

    A, B, AB, BA = separate_output_values(Y, D, N, calc_second_order)
    r = rng(N, size=(N, num_resamples))
    Z = norm.ppf(0.5 + conf_level / 2)

    if not parallel:
        S = create_Si_dict(D, num_resamples, keep_resamples, calc_second_order)

        for j in range(D):
            S["S1"][j] = first_order(A, AB[:, j], B)
            S1_conf_j = first_order(A[r], AB[r, j], B[r])

            if keep_resamples:
                S["S1_conf_all"][:, j] = S1_conf_j

            var_diff = np.r_[A[r], B[r]].ptp()
            if var_diff != 0.0:
                S["S1_conf"][j] = Z * S1_conf_j.std(ddof=1)
            else:
                S["S1_conf"][j] = 0.0

            S["ST"][j] = total_order(A, AB[:, j], B)

            ST_conf_j = total_order(A[r], AB[r, j], B[r])

            if keep_resamples:
                S["ST_conf_all"][:, j] = ST_conf_j

            if var_diff != 0.0:
                S["ST_conf"][j] = Z * ST_conf_j.std(ddof=1)
            else:
                S["ST_conf"][j] = 0.0

        # Second order (+conf.)
        if calc_second_order:
            for j in range(D):
                for k in range(j + 1, D):
                    S["S2"][j, k] = second_order(A, AB[:, j], AB[:, k], BA[:, j], B)
                    S["S2_conf"][j, k] = Z * second_order(
                        A[r], AB[r, j], AB[r, k], BA[r, j], B[r]
                    ).std(ddof=1)
    else:
        tasks, n_processors = create_task_list(D, calc_second_order, n_processors)

        func = partial(sobol_parallel, Z, A, AB, BA, B, r)
        pool = Pool(n_processors)
        S_list = pool.map_async(func, tasks)
        pool.close()
        pool.join()

        S = Si_list_to_dict(
            S_list.get(), D, num_resamples, keep_resamples, calc_second_order
        )

    # Add problem context and override conversion method for special case
    S.problem = problem
    S.to_df = MethodType(to_df, S)

    # Print results to console
    if print_to_console:
        res = S.to_df()
        for df in res:
            print(df)

    return S


def first_order(A, AB, B):
    """
    First order estimator following Saltelli et al. 2010 CPC, normalized by
    sample variance
    """
    y = np.r_[A, B]
    if y.ptp() == 0:
        warn(CONST_RESULT_MSG)
        return np.array([0.0])

    return np.mean(B * (AB - A), axis=0) / np.var(y, axis=0)


def total_order(A, AB, B):
    """
    Total order estimator following Saltelli et al. 2010 CPC, normalized by
    sample variance
    """
    y = np.r_[A, B]
    if y.ptp() == 0:
        warn(CONST_RESULT_MSG)
        return np.array([0.0])

    return 0.5 * np.mean((A - AB) ** 2, axis=0) / np.var(y, axis=0)


def second_order(A, ABj, ABk, BAj, B):
    """Second order estimator following Saltelli 2002"""
    y = np.r_[A, B]
    if y.ptp() == 0:
        warn(CONST_RESULT_MSG)
        return np.array([0.0])

    Vjk = np.mean(BAj * ABk - A * B, axis=0) / np.var(y, axis=0)
    Sj = first_order(A, ABj, B)
    Sk = first_order(A, ABk, B)

    return Vjk - Sj - Sk


def create_Si_dict(
    D: int, num_resamples: int, keep_resamples: bool, calc_second_order: bool
):
    """initialize empty dict to store sensitivity indices"""
    S = ResultDict((k, np.zeros(D)) for k in ("S1", "S1_conf", "ST", "ST_conf"))

    if keep_resamples:
        # Create entries to store intermediate resampling results
        S["S1_conf_all"] = np.zeros((num_resamples, D))
        S["ST_conf_all"] = np.zeros((num_resamples, D))

    if calc_second_order:
        S["S2"] = np.full((D, D), np.nan)
        S["S2_conf"] = np.full((D, D), np.nan)

    return S


def separate_output_values(Y, D, N, calc_second_order):
    AB = np.zeros((N, D))
    BA = np.zeros((N, D)) if calc_second_order else None
    step = 2 * D + 2 if calc_second_order else D + 2

    A = Y[0 : Y.size : step]
    B = Y[(step - 1) : Y.size : step]
    for j in range(D):
        AB[:, j] = Y[(j + 1) : Y.size : step]
        if calc_second_order:
            BA[:, j] = Y[(j + 1 + D) : Y.size : step]

    return A, B, AB, BA


def sobol_parallel(Z, A, AB, BA, B, r, tasks):
    sobol_indices = []
    for d, j, k in tasks:
        if d == "S1":
            s = first_order(A, AB[:, j], B)
        elif d == "S1_conf":
            s = Z * first_order(A[r], AB[r, j], B[r]).std(ddof=1)
        elif d == "ST":
            s = total_order(A, AB[:, j], B)
        elif d == "ST_conf":
            s = Z * total_order(A[r], AB[r, j], B[r]).std(ddof=1)
        elif d == "S2":
            s = second_order(A, AB[:, j], AB[:, k], BA[:, j], B)
        elif d == "S2_conf":
            s = Z * second_order(A[r], AB[r, j], AB[r, k], BA[r, j], B[r]).std(ddof=1)
        sobol_indices.append([d, j, k, s])

    return sobol_indices


def create_task_list(D, calc_second_order, n_processors):
    """
    Create list with one entry (key, parameter 1, parameter 2) per sobol
    index (+conf.). This is used to supply parallel tasks to
    multiprocessing.Pool
    """
    tasks_first_order = [
        [d, j, None] for j in range(D) for d in ("S1", "S1_conf", "ST", "ST_conf")
    ]

    # Add second order (+conf.) to tasks
    tasks_second_order = []
    if calc_second_order:
        tasks_second_order = [
            [d, j, k]
            for j in range(D)
            for k in range(j + 1, D)
            for d in ("S2", "S2_conf")
        ]

    if n_processors is None:
        n_processors = min(
            cpu_count(), len(tasks_first_order) + len(tasks_second_order)
        )

    if not calc_second_order:
        tasks = np.array_split(tasks_first_order, n_processors)
    else:
        # merges both lists alternating its elements and splits the
        # resulting lists into n_processors sublists
        tasks = np.array_split(
            [
                v
                for v in sum(
                    zip_longest(tasks_first_order[::-1], tasks_second_order), ()
                )
                if v is not None
            ],
            n_processors,
        )

    return tasks, n_processors


def Si_list_to_dict(
    S_list, D: int, num_resamples: int, keep_resamples: bool, calc_second_order: bool
):
    """Convert the parallel output into the regular dict format for
    printing/returning"""
    S = create_Si_dict(D, num_resamples, keep_resamples, calc_second_order)
    L = []
    for list in S_list:  # first reformat to flatten
        L += list

    for s in L:  # First order (+conf.)
        if s[2] is None:
            S[s[0]][s[1]] = s[3]
        else:
            S[s[0]][s[1], s[2]] = s[3]

    return S


def Si_to_pandas_dict(S_dict):
    """Convert Si information into Pandas DataFrame compatible dict.


    Examples
    --------
        >>> X = saltelli.sample(problem, 512)
        >>> Y = Ishigami.evaluate(X)
        >>> Si = sobol.analyze(problem, Y, print_to_console=True)
        >>> T_Si, first_Si, (idx, second_Si) = sobol.Si_to_pandas_dict(Si, problem)


    Parameters
    ----------
    S_dict : ResultDict
        Sobol sensitivity indices

    See Also
    --------
    Si_list_to_dict

    Returns
    -------
    tuple : of total, first, and second order sensitivities.
            Total and first order are dicts.
            Second order sensitivities contain a tuple of parameter name
            combinations for use as the DataFrame index and second order
            sensitivities.
            If no second order indices found, then returns tuple of
            (None, None)
    """
    problem = S_dict.problem
    total_order = {"ST": S_dict["ST"], "ST_conf": S_dict["ST_conf"]}
    first_order = {"S1": S_dict["S1"], "S1_conf": S_dict["S1_conf"]}

    idx = None
    second_order = None
    if "S2" in S_dict:
        groups = _check_groups(problem)
        if groups:
            names, _ = extract_group_names(groups)
        else:
            names = problem.get("names")

        if len(names) > 2:
            idx = list(combinations(names, 2))
        else:
            idx = (names,)

        second_order = {
            "S2": [S_dict["S2"][names.index(i[0]), names.index(i[1])] for i in idx],
            "S2_conf": [
                S_dict["S2_conf"][names.index(i[0]), names.index(i[1])] for i in idx
            ],
        }
    return total_order, first_order, (idx, second_order)


def to_df(self):
    """Conversion method to Pandas DataFrame. To be attached to ResultDict.

    Returns
    -------
    List : of Pandas DataFrames in order of Total, First, Second

    Example
    -------
    >>> Si = sobol.analyze(problem, Y, print_to_console=True)
    >>> total_Si, first_Si, second_Si = Si.to_df()
    """
    total, first, (idx, second) = Si_to_pandas_dict(self)

    problem = self.problem
    groups = _check_groups(problem)
    if not groups:
        names = problem.get("names")
    else:
        names, _ = extract_group_names(groups)

    ret = [pd.DataFrame(total, index=names), pd.DataFrame(first, index=names)]

    if second:
        ret += [pd.DataFrame(second, index=idx)]

    return ret


def cli_parse(parser):
    parser.add_argument(
        "--max-order",
        type=int,
        required=False,
        default=2,
        choices=[1, 2],
        help="Maximum order of sensitivity indices to " "calculate",
    )
    parser.add_argument(
        "-r",
        "--resamples",
        type=int,
        required=False,
        default=100,
        help="Number of bootstrap resamples for Sobol " "confidence intervals",
    )
    parser.add_argument(
        "--parallel",
        action="store_true",
        help="Makes " "use of parallelization.",
        dest="parallel",
    )
    parser.add_argument(
        "--processors",
        type=int,
        required=False,
        default=None,
        help="Number of processors to be used with the " + "parallel option.",
        dest="n_processors",
    )
    return parser


def cli_action(args):
    problem = read_param_file(args.paramfile)
    Y = np.loadtxt(
        args.model_output_file, delimiter=args.delimiter, usecols=(args.column,)
    )

    analyze(
        problem,
        Y,
        (args.max_order == 2),
        num_resamples=args.resamples,
        print_to_console=True,
        parallel=args.parallel,
        n_processors=args.n_processors,
        seed=args.seed,
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



import numpy as np


def evaluate(X: np.ndarray, A: float = 7.0, B: float = 0.1) -> np.ndarray:
    r"""Non-monotonic Ishigami-Homma three parameter test function:

    `f(x) = \sin(x_{1}) + A \sin(x_{2})^2 + Bx^{4}_{3}\sin(x_{1})`

    This test function is commonly used to benchmark global sensitivity
    methods as variance-based sensitivities of this function can be
    analytically determined.

    See listed references below.

    In [2], the expected first-order indices are:

        x1: 0.3139
        x2: 0.4424
        x3: 0.0

    when A = 7, B = 0.1 when conducting Sobol' analysis with the
    Saltelli sampling method with a sample size of 1000.


    Parameters
    ----------
    X : np.ndarray
        An `N*D` array holding values for each parameter, where `N` is the
        number of samples and `D` is the number of parameters
        (in this case, three).
    A : float
        Constant `A` parameter
    B : float
        Constant `B` parameter

    Returns
    -------
    Y : np.ndarray

    References
    ----------
    .. [1] Ishigami, T., Homma, T., 1990.
           An importance quantification technique in uncertainty analysis for
           computer models.
           Proceedings. First International Symposium on Uncertainty Modeling
           and Analysis.
           https://doi.org/10.1109/ISUMA.1990.151285

    .. [2] Saltelli, A., Ratto, M., Andres, T., Campolongo, F., Cariboni, J.,
           Gatelli, D., Saisana, M., Tarantola, S., 2008.
           Global Sensitivity Analysis: The Primer. Wiley, West Sussex, U.K.
           https://dx.doi.org/10.1002/9780470725184
    """
    Y = np.zeros(X.shape[0])
    Y = (
        np.sin(X[:, 0])
        + A * np.power(np.sin(X[:, 1]), 2)
        + B * np.power(X[:, 2], 4) * np.sin(X[:, 0])
    )

    return Y



from typing import Dict

import numpy as np
from scipy.stats import ks_2samp

from . import common_args
from ..util import read_param_file, ResultDict, extract_group_names, _check_groups


def analyze(
    problem: Dict,
    X: np.ndarray,
    Y: np.ndarray,
    S: int = 10,
    print_to_console: bool = False,
    seed: int = None,
):
    """Performs PAWN sensitivity analysis.

    The PAWN method [1] is a moment-independent approach to Global Sensitivity
    Analysis (GSA). It is described as producing robust results at relatively
    low sample sizes (see [2]) for the purpose of factor ranking and screening.

    The distribution of model outputs is examined rather than
    their variation as is typical in other common GSA approaches. The PAWN
    method further distinguishes itself from other moment-independent
    approaches by characterizing outputs by their cumulative distribution
    function (CDF) as opposed to their probability distribution function.
    As the CDF for a given random variable is typically normally distributed,
    PAWN can be more appropriately applied when outputs are highly-skewed or
    multi-modal, for which variance-based methods may produce unreliable
    results.

    PAWN characterizes the relationship between inputs and outputs by
    quantifying the variation in the output distributions after conditioning
    an input. A factor is deemed non-influential if distributions coincide at
    all ``S`` conditioning intervals. The Kolmogorov-Smirnov statistic is used
    as a measure of distance between the distributions.

    This implementation reports the PAWN index at the min, mean, median, and
    max across the slides/conditioning intervals as well as the coefficient of
    variation (``CV``). The median value is the typically reported value. As
    the ``CV`` is (standard deviation / mean), it indicates the level of
    variability across the slides, with values closer to zero indicating lower
    variation.


    Notes
    -----
    Compatible with:
        all samplers

    This implementation ignores all NaNs.

    When applied to grouped factors, the analysis is conducted on each factor
    individually, and the mean of their results are reported.


    Examples
    --------
        >>> X = latin.sample(problem, 1000)
        >>> Y = Ishigami.evaluate(X)
        >>> Si = pawn.analyze(problem, X, Y, S=10, print_to_console=False)


    Parameters
    ----------
    problem : dict
        The problem definition
    X : numpy.array
        A NumPy array containing the model inputs
    Y : numpy.array
        A NumPy array containing the model outputs
    S : int
        Number of slides; the conditioning intervals (default 10)
    print_to_console : bool
        Print results directly to console (default False)
    seed : int
        Seed value to ensure deterministic results


    References
    ----------
    1. Pianosi, F., Wagener, T., 2015.
           A simple and efficient method for global sensitivity analysis
           based on cumulative distribution functions.
           Environmental Modelling & Software 67, 1-11.
           https://doi.org/10.1016/j.envsoft.2015.01.004

    2. Pianosi, F., Wagener, T., 2018.
           Distribution-based sensitivity analysis from a generic input-output sample.
           Environmental Modelling & Software 108, 197-207.
           https://doi.org/10.1016/j.envsoft.2018.07.019

    3. Baroni, G., Francke, T., 2020.
           An effective strategy for combining variance- and
           distribution-based global sensitivity analysis.
           Environmental Modelling & Software, 134, 104851.
           https://doi.org/10.1016/j.envsoft.2020.104851

    4. Baroni, G., Francke, T., 2020.
           GSA-cvd
           Combining variance- and distribution-based global sensitivity analysis
           https://github.com/baronig/GSA-cvd
    """
    if seed:
        np.random.seed(seed)

    D = problem["num_vars"]
    groups = _check_groups(problem)
    if not groups:
        var_names = problem["names"]
    else:
        var_names, _ = extract_group_names(problem.get("groups", []))

    results = np.full((D, 5), np.nan)
    temp_pawn = np.full((S, D), np.nan)

    step = 1 / S
    for d_i in range(D):
        seq = np.arange(0, 1 + step, step)
        X_di = X[:, d_i]
        X_q = np.nanquantile(X_di, seq)

        for s in range(S):
            Y_sel = Y[(X_di >= X_q[s]) & (X_di < X_q[s + 1])]
            if len(Y_sel) == 0:
                # no available samples
                continue

            # KD value
            # Function returns a KS object which holds the KS statistic
            # and p-value
            # Note from scipy documentation:
            # if the K-S statistic is small or the p-value is high, then
            # we cannot reject the hypothesis that the distributions of
            # the two samples are the same.
            ks = ks_2samp(Y_sel, Y)
            temp_pawn[s, d_i] = ks.statistic

        p_ind = temp_pawn[:, d_i]
        mins = np.nanmin(p_ind)
        mean = np.nanmean(p_ind)
        med = np.nanmedian(p_ind)
        maxs = np.nanmax(p_ind)
        cv = np.nanstd(p_ind) / mean
        results[d_i, :] = [mins, mean, med, maxs, cv]

    if groups:
        groups = np.array(groups)
        unique_grps = [*dict.fromkeys(groups)]
        tmp = np.full((len(unique_grps), 5), np.nan)

        # Take the mean of effects from parameters that are grouped together
        for grp_id, grp in enumerate(unique_grps):
            tmp[grp_id, :] = np.mean(results[groups == grp, :], axis=0)

        results = tmp
        tmp = None

    Si = ResultDict(
        [
            ("minimum", results[:, 0]),
            ("mean", results[:, 1]),
            ("median", results[:, 2]),
            ("maximum", results[:, 3]),
            ("CV", results[:, 4]),
        ]
    )
    Si["names"] = var_names

    if print_to_console:
        print(Si.to_df())

    return Si


def cli_parse(parser):
    parser.add_argument(
        "-X", "--model-input-file", type=str, required=True, help="Model input file"
    )

    parser.add_argument(
        "-S", "--slices", type=int, required=False, help="Number of slices to take"
    )
    return parser


def cli_action(args):
    problem = read_param_file(args.paramfile)
    X = np.loadtxt(args.model_input_file, delimiter=args.delimiter)
    Y = np.loadtxt(
        args.model_output_file, delimiter=args.delimiter, usecols=(args.column,)
    )
    analyze(problem, X, Y, S=args.slices, print_to_console=True, seed=args.seed)


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



"""
Defines a family of algorithms for generating samples

The sample a for use with :class:`SALib.analyze.morris.analyze`,
encapsulate each one, and makes them interchangeable.

Example
-------
>>> localoptimisation = LocalOptimisation()
>>> context = SampleMorris(localoptimisation)
>>> context.sample(input_sample, num_samples, num_params, k_choices, groups)
"""
import abc

import numpy as np  # type: ignore
from scipy.spatial.distance import cdist  # type: ignore


class SampleMorris(object):
    """Computes the optimum `k_choices` of trajectories from the input_sample.

    Parameters
    ----------
    strategy : :class:`Strategy`
    """

    def __init__(self, strategy):
        self._strategy = strategy

    def sample(self, input_sample, num_samples, num_params, k_choices, num_groups):
        """Computes the optimum k_choices of trajectories
        from the input_sample.

        Parameters
        ----------
        input_sample : numpy.ndarray
        num_samples : int
            The number of samples to generate
        num_params : int
            The number of parameters
        k_choices : int
            The number of optimal trajectories
        num_groups : int
            The number of groups

        Returns
        -------
        numpy.ndarray
            An array of optimal trajectories
        """
        return self._strategy.sample(
            input_sample, num_samples, num_params, k_choices, num_groups
        )


class Strategy:
    """
    Declare an interface common to all supported algorithms.
    :class:`SampleMorris` uses this interface to call the algorithm
    defined by a ConcreteStrategy.
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def _sample(self, input_sample, num_samples, num_params, k_choices, num_groups):
        """Implement this in your class

        Parameters
        ----------
        input_sample : numpy.ndarray
        num_samples : int
            The number of samples to generate
        num_params : int
            The number of parameters
        k_choices : int
            The number of optimal trajectories
        num_groups : int
            The number of groups

        Returns
        -------
        list
            A list of trajectory indices
        """
        pass

    def sample(self, input_sample, num_samples, num_params, k_choices, num_groups=None):
        """Computes the optimum k_choices of trajectories
        from the input_sample.

        Parameters
        ----------
        input_sample : numpy.ndarray
        num_samples : int
            The number of samples to generate
        num_params : int
            The number of parameters
        k_choices : int
            The number of optimal trajectories
        num_groups : int, default=None
            The number of groups

        Returns
        -------
        numpy.ndarray
        """
        self.run_checks(num_samples, k_choices)
        maximum_combo = self._sample(
            input_sample, num_samples, num_params, k_choices, num_groups
        )

        assert isinstance(maximum_combo, list)

        output = self.compile_output(
            input_sample, num_samples, num_params, maximum_combo, num_groups
        )
        return output

    @staticmethod
    def run_checks(number_samples, k_choices):
        """Runs checks on `k_choices`"""
        assert isinstance(
            k_choices, int
        ), "Number of optimal trajectories should be an integer"

        if k_choices < 2:
            raise ValueError(
                "The number of optimal trajectories must be set to 2 or more."
            )
        if k_choices >= number_samples:
            msg = "The number of optimal trajectories should be less than the \
                    number of samples"
            raise ValueError(msg)

    @staticmethod
    def _make_index_list(num_samples, num_params, num_groups=None):
        """Identify indices of input sample associated with each trajectory

        For each trajectory, identifies the indexes of the input sample which
        is a function of the number of factors/groups and the number of samples

        Parameters
        ----------
        num_samples : int
            The number of trajectories
        num_params : int
            The number of parameters
        num_groups : int
            The number of groups

        Returns
        -------
        list of numpy.ndarray

        Example
        -------
        >>> BruteForce()._make_index_list(num_samples=4, num_params=3,
            num_groups=2)
        [np.array([0, 1, 2]), np.array([3, 4, 5]), np.array([6, 7, 8]),
         np.array([9, 10, 11])]
        """
        if num_groups is None:
            num_groups = num_params

        index_list = []
        for j in range(num_samples):
            index_list.append(np.arange(num_groups + 1) + j * (num_groups + 1))
        return index_list

    def compile_output(
        self, input_sample, num_samples, num_params, maximum_combo, num_groups=None
    ):
        """Picks the trajectories from the input

        Parameters
        ----------
        input_sample : numpy.ndarray
        num_samples : int
        num_params : int
        maximum_combo : list
        num_groups : int

        """

        if num_groups is None:
            num_groups = num_params

        self.check_input_sample(input_sample, num_groups, num_samples)

        index_list = self._make_index_list(num_samples, num_params, num_groups)

        output = np.zeros((np.size(maximum_combo) * (num_groups + 1), num_params))
        for counter, combo in enumerate(maximum_combo):
            output[index_list[counter]] = np.array(input_sample[index_list[combo]])
        return output

    @staticmethod
    def check_input_sample(input_sample, num_params, num_samples):
        """Check the `input_sample` is valid

        Checks input sample is:
            - the correct size
            - values between 0 and 1

        Parameters
        ----------
        input_sample : numpy.ndarray
        num_params : int
        num_samples : int
        """
        assert type(input_sample) == np.ndarray, "Input sample is not an numpy array"
        assert (
            input_sample.shape[0] == (num_params + 1) * num_samples
        ), "Input sample does not match number of parameters or groups"
        assert np.any(
            (input_sample >= 0) | (input_sample <= 1)
        ), "Input sample must be scaled between 0 and 1"

    @staticmethod
    def compute_distance(m, l):  # noqa: E741
        """Compute distance between two trajectories

        Parameters
        ----------
        m : np.ndarray

        l : np.ndarray

        Returns
        -------
        numpy.ndarray
        """

        if np.shape(m) != np.shape(l):
            raise ValueError("Input matrices are different sizes")
        if np.array_equal(m, l):
            # print("Trajectory %s and %s are equal" % (m, l))
            distance = 0
        else:
            distance = np.array(np.sum(cdist(m, l)), dtype=np.float32)

        return distance

    def compute_distance_matrix(
        self,
        input_sample,
        num_samples,
        num_params,
        num_groups=None,
        local_optimization=False,
    ):
        """Computes the distance between each and every trajectory

        Each entry in the matrix represents the sum of the geometric distances
        between all the pairs of points of the two trajectories

        If the `groups` argument is filled, then the distances are still
        calculated for each trajectory,

        Parameters
        ----------
        input_sample : numpy.ndarray
            The input sample of trajectories for which to compute
            the distance matrix
        num_samples : int
            The number of trajectories
        num_params : int
            The number of factors
        num_groups : int, default=None
            The number of groups
        local_optimization : bool, default=False
            If True, fills the lower triangle of the distance matrix

        Returns
        -------
        distance_matrix : numpy.ndarray

        """
        if num_groups:
            self.check_input_sample(input_sample, num_groups, num_samples)
        else:
            self.check_input_sample(input_sample, num_params, num_samples)

        index_list = self._make_index_list(num_samples, num_params, num_groups)
        distance_matrix = np.zeros((num_samples, num_samples), dtype=np.float32)

        for j in range(num_samples):
            input_1 = input_sample[index_list[j]]
            for k in range(j + 1, num_samples):
                input_2 = input_sample[index_list[k]]

                # Fills the lower triangle of the matrix
                if local_optimization is True:
                    distance_matrix[j, k] = self.compute_distance(input_1, input_2)

                distance_matrix[k, j] = self.compute_distance(input_1, input_2)
        return distance_matrix



import argparse


def setup(parser):
    parser = argparse.ArgumentParser(
        description="Perform sensitivity analysis on model output"
    )
    parser.add_argument(
        "-p", "--paramfile", type=str, required=True, help="Parameter range file"
    )
    parser.add_argument(
        "-Y", "--model-output-file", type=str, required=True, help="Model output file"
    )
    parser.add_argument(
        "-c",
        "--column",
        type=int,
        required=False,
        default=0,
        help="Column of output to analyze",
    )
    parser.add_argument(
        "--delimiter",
        type=str,
        required=False,
        default=" ",
        help="Column delimiter in model output file",
    )
    parser.add_argument(
        "-s", "--seed", type=int, required=False, default=None, help="Random Seed"
    )
    return parser


def create(cli_parser=None):
    parser = argparse.ArgumentParser(
        description="Perform sensitivity analysis on model output"
    )
    parser = setup(parser)

    if cli_parser:
        parser = cli_parser(parser)

    return parser


def run_cli(cli_parser, run_analysis, known_args=None):
    parser = create(cli_parser)
    args = parser.parse_args(known_args)

    run_analysis(args)



import warnings
from typing import Dict, Optional, Union

import numpy as np
from scipy.stats import qmc

from . import common_args
from ..util import scale_samples, read_param_file, compute_groups_matrix, _check_groups


def sample(
    problem: Dict,
    N: int,
    *,
    calc_second_order: bool = True,
    scramble: bool = True,
    skip_values: int = 0,
    seed: Optional[Union[int, np.random.Generator]] = None,
):
    """Generates model inputs using Saltelli's extension of the Sobol' sequence.

    The Sobol' sequence is a popular quasi-random low-discrepancy sequence used
    to generate uniform samples of parameter space.

    Returns a NumPy matrix containing the model inputs using Saltelli's
    sampling scheme.
    Saltelli's scheme extends the Sobol' sequence in a way to reduce
    the error rates in the resulting sensitivity index calculations. If
    `calc_second_order` is False, the resulting matrix has ``N * (D + 2)``
    rows, where ``D`` is the number of parameters.
    If `calc_second_order` is `True`, the resulting matrix has ``N * (2D + 2)``
    rows.

    These model inputs are intended to be used with
    :func:`SALib.analyze.sobol.analyze`.
	
	Notes
    -----
    The initial points of the Sobol' sequence has some repetition (see Table 2
    in Campolongo [1]__), which can be avoided by scrambling the sequence.

    Another option, not recommended and available for educational purposes,
    is to use the `skip_values` parameter.
    Skipping values reportedly improves the uniformity of samples.
    But, it has been shown that naively skipping values may reduce accuracy,
    increasing the number of samples needed to achieve convergence
    (see Owen [2]__).

	Parameters
    ----------
    problem : dict, 
        The problem definition.
    N : int
        The number of samples to generate.
        Ideally a power of 2 and <= `skip_values`.
    calc_second_order : bool, optional
        Calculate second-order sensitivities. Default is True.
    scramble : bool, optional
        If True, use LMS+shift scrambling. Otherwise, no scrambling is done.
        Default is True.
    skip_values : int, optional
        Number of points in Sobol' sequence to skip, ideally a value of base 2.
        It's recommended not to change this value and use `scramble` instead.
        `scramble` and `skip_values` can be used together.
        Default is 0.
    seed : {None, int, `numpy.random.Generator`}, optional
        If `seed` is None the `numpy.random.Generator` generator is used.
        If `seed` is an int, a new ``Generator`` instance is used,
        seeded with `seed`.
        If `seed` is already a ``Generator`` instance then that instance is
        used. Default is None.

    References
    ----------
    1. Campolongo, F., Saltelli, A., Cariboni, J., 2011.
       From screening to quantitative sensitivity analysis.
       A unified approach.
       Computer Physics Communications 182, 978-988.
       https://doi.org/10.1016/j.cpc.2010.12.039

    2. Owen, A. B., 2020.
       On dropping the first Sobol' point.
       arXiv:2008.08051 [cs, math, stat].
       Available at: http://arxiv.org/abs/2008.08051
       (Accessed: 20 April 2021).

    3. Saltelli, A., 2002.
       Making best use of model evaluations to compute sensitivity indices.
       Computer Physics Communications 145, 280-297.
       https://doi.org/10.1016/S0010-4655(02)00280-1

    4. Sobol', I.M., 2001.
       Global sensitivity indices for nonlinear mathematical models and
       their Monte Carlo estimates.
       Mathematics and Computers in Simulation,
       The Second IMACS Seminar on Monte Carlo Methods 55, 271-280.
       https://doi.org/10.1016/S0378-4754(00)00270-6
    """
    D = problem["num_vars"]
    groups = _check_groups(problem)

    # Create base sequence - could be any type of sampling
    qrng = qmc.Sobol(d=2 * D, scramble=scramble, seed=seed)

    # fast-forward logic
    if skip_values > 0 and isinstance(skip_values, int):
        M = skip_values
        if not ((M & (M - 1) == 0) and (M != 0 and M - 1 != 0)):
            msg = f"""
            Convergence properties of the Sobol' sequence is only valid if
            `skip_values` ({M}) is a power of 2.
            """
            warnings.warn(msg, stacklevel=2)

        # warning when N > skip_values
        # see https://github.com/scipy/scipy/pull/10844#issuecomment-673029539
        n_exp = int(np.log2(N))
        m_exp = int(np.log2(M))
        if n_exp > m_exp:
            msg = (
                "Convergence may not be valid as the number of "
                "requested samples is"
                f" > `skip_values` ({N} > {M})."
            )
            warnings.warn(msg, stacklevel=2)

        qrng.fast_forward(M)
    elif skip_values < 0 or not isinstance(skip_values, int):
        raise ValueError("`skip_values` must be a positive integer.")

    # sample Sobol' sequence
    base_sequence = qrng.random(N)

    if not groups:
        Dg = problem["num_vars"]
    else:
        G, group_names = compute_groups_matrix(groups)
        Dg = len(set(group_names))

    if calc_second_order:
        saltelli_sequence = np.zeros([(2 * Dg + 2) * N, D])
    else:
        saltelli_sequence = np.zeros([(Dg + 2) * N, D])

    index = 0

    for i in range(N):
        # Copy matrix "A"
        for j in range(D):
            saltelli_sequence[index, j] = base_sequence[i, j]

        index += 1

        # Cross-sample elements of "B" into "A"
        for k in range(Dg):
            for j in range(D):
                if (not groups and j == k) or (groups and group_names[k] == groups[j]):
                    saltelli_sequence[index, j] = base_sequence[i, j + D]
                else:
                    saltelli_sequence[index, j] = base_sequence[i, j]

            index += 1

        # Cross-sample elements of "A" into "B"
        # Only needed if you're doing second-order indices (true by default)
        if calc_second_order:
            for k in range(Dg):
                for j in range(D):
                    if (not groups and j == k) or (
                        groups and group_names[k] == groups[j]
                    ):
                        saltelli_sequence[index, j] = base_sequence[i, j]
                    else:
                        saltelli_sequence[index, j] = base_sequence[i, j + D]

                index += 1

        # Copy matrix "B"
        for j in range(D):
            saltelli_sequence[index, j] = base_sequence[i, j + D]

        index += 1

    saltelli_sequence = scale_samples(saltelli_sequence, problem)
    return saltelli_sequence


def cli_parse(parser):
    """Add method specific options to CLI parser.

    Parameters
    ----------
    parser : argparse object

    Returns
    -------
    Updated argparse object
    """
    parser.add_argument(
        "--max-order",
        type=int,
        required=False,
        default=2,
        choices=[1, 2],
        help="Maximum order of sensitivity indices to calculate",
    )

    parser.add_argument(
        "--scramble",
        type=int,
        required=False,
        default=True,
        help="Use scrambled sequence",
    )

    parser.add_argument(
        "--skip-values",
        type=int,
        required=False,
        default=None,
        help="Number of sample points to skip (default: next largest power of"
        " 2 from `samples`). Not recommended (use `scramble` instead).",
    )

    return parser


def cli_action(args):
    """Run sampling method

    Parameters
    ----------
    args : argparse namespace
    """
    problem = read_param_file(args.paramfile)
    param_values = sample(
        problem,
        args.samples,
        calc_second_order=(args.max_order == 2),
        scramble=args.scramble,
    )
    np.savetxt(
        args.output,
        param_values,
        delimiter=args.delimiter,
        fmt="%." + str(args.precision) + "e",
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



#!/usr/bin/env python
# coding=utf8

import numpy as np
from scipy.signal import periodogram
from scipy.stats import norm

from . import common_args
from ..util import read_param_file, ResultDict


def analyze(
    problem,
    X,
    Y,
    M=10,
    num_resamples=100,
    conf_level=0.95,
    print_to_console=False,
    seed=None,
):
    """Performs the Random Balanced Design - Fourier Amplitude Sensitivity Test
    (RBD-FAST) on model outputs.

    Returns a dictionary with keys 'S1', where each entry is a list of
    size D (the number of parameters) containing the indices in the same order
    as the parameter file.

    Notes
    -----
    Compatible with:
        all samplers


    Examples
    --------
        >>> X = latin.sample(problem, 1000)
        >>> Y = Ishigami.evaluate(X)
        >>> Si = rbd_fast.analyze(problem, X, Y, print_to_console=False)


    Parameters
    ----------
    problem : dict
        The problem definition
    X : numpy.array
        A NumPy array containing the model inputs
    Y : numpy.array
        A NumPy array containing the model outputs
    M : int
        The interference parameter, i.e., the number of harmonics to sum in
        the Fourier series decomposition (default 10)
    print_to_console : bool
        Print results directly to console (default False)
    seed : int
        Seed to generate a random number


    References
    ----------
    1. S. Tarantola, D. Gatelli and T. Mara (2006)
       Random Balance Designs for the Estimation of First Order Global
       Sensitivity Indices,
       Reliability Engineering and System Safety, 91:6, 717-727
       https://doi.org/10.1016/j.ress.2005.06.003

    2. Elmar Plischke (2010)
        An effective algorithm for computing global sensitivity indices 
        (EASI),
        Reliability Engineering & System Safety,
        95:4, 354-360. doi:10.1016/j.ress.2009.11.005

    3. Jean-Yves Tissot, Clémentine Prieur (2012)
        Bias correction for the estimation of sensitivity indices based
        on random balance designs,
        Reliability Engineering and System Safety, Elsevier, 107, 205-213.
        doi:10.1016/j.ress.2012.06.010

    4. Jeanne Goffart, Mickael Rabouille & Nathan Mendes (2015)
         Uncertainty and sensitivity analysis applied to hygrothermal
         simulation of a brick building in a hot and humid climate,
         Journal of Building Performance Simulation.
         doi:10.1080/19401493.2015.1112430
    """
    if seed:
        np.random.seed(seed)

    D = problem["num_vars"]
    N = Y.size

    # Calculate and Output the First Order Value
    Si = ResultDict((k, [None] * D) for k in ["S1", "S1_conf"])
    Si["names"] = problem["names"]

    for i in range(D):
        S1 = compute_first_order(permute_outputs(X[:, i], Y), M)
        S1 = unskew_S1(S1, M, N)
        Si["S1"][i] = S1
        Si["S1_conf"][i] = bootstrap(X[:, i], Y, M, num_resamples, conf_level)

    if print_to_console:
        print(Si.to_df())

    return Si


def permute_outputs(X, Y):
    """
    Permute the output according to one of the inputs as in [_2]

    References
    ----------
    .. [2] Elmar Plischke (2010) "An effective algorithm for computing global
          sensitivity indices (EASI) Reliability Engineering & System Safety",
          95:4, 354-360. doi:10.1016/j.ress.2009.11.005

    """
    permutation_index = np.argsort(X)
    permutation_index = np.concatenate(
        [permutation_index[::2], permutation_index[1::2][::-1]]
    )
    return Y[permutation_index]


def compute_first_order(permuted_outputs, M):
    _, Pxx = periodogram(permuted_outputs)
    V = np.sum(Pxx[1:])
    D1 = np.sum(Pxx[1 : M + 1])
    return D1 / V


def unskew_S1(S1, M, N):
    """
    Unskew the sensivity indice
    (Jean-Yves Tissot, Clémentine Prieur (2012) "Bias correction for the
    estimation of sensitivity indices based on random balance designs.",
    Reliability Engineering and System Safety, Elsevier, 107, 205-213.
    doi:10.1016/j.ress.2012.06.010)
    """
    lamb = (2 * M) / N
    return S1 - lamb / (1 - lamb) * (1 - S1)


def bootstrap(X_d, Y, M, resamples, conf_level):
    # Use half of available data each time
    T_data = X_d.shape[0]
    n_size = int(T_data * 0.5)

    res = np.zeros(resamples)
    for i in range(resamples):
        sample_idx = np.random.choice(T_data, replace=True, size=n_size)
        X_rs, Y_rs = X_d[sample_idx], Y[sample_idx]
        S1 = compute_first_order(permute_outputs(X_rs, Y_rs), M)
        S1 = unskew_S1(S1, M, Y_rs.size)
        res[i] = S1

    return norm.ppf(0.5 + conf_level / 2.0) * res.std(ddof=1)


def cli_parse(parser):
    parser.add_argument(
        "-X", "--model-input-file", type=str, required=True, help="Model input file"
    )
    parser.add_argument(
        "-M", "--M", type=int, required=False, default=10, help="Inference parameter"
    )
    parser.add_argument(
        "-r",
        "--resamples",
        type=int,
        required=False,
        default=100,
        help="Number of bootstrap resamples for Sobol " "confidence intervals",
    )
    return parser


def cli_action(args):
    problem = read_param_file(args.paramfile)
    X = np.loadtxt(args.model_input_file, delimiter=args.delimiter)
    Y = np.loadtxt(
        args.model_output_file, delimiter=args.delimiter, usecols=(args.column,)
    )
    analyze(
        problem,
        X,
        Y,
        M=args.M,
        num_resamples=args.resamples,
        print_to_console=True,
        seed=args.seed,
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



import numpy as np
import matplotlib.pyplot as plt


__all__ = ["heatmap"]


# magic string indicating DF columns holding conf bound values
CONF_COLUMN = "_conf"


def heatmap(sp, metric: str, index: str, title: str = None, ax=None):
    """Plot a heatmap of the target metric.

    Parameters
    ----------
    sp : object, SALib ProblemSpec
    metric : str, metric to plot. Defaults to first metric/result output if `None`.
    index : str, sensitivity indices to plot ('S1', 'ST', etc). Displays all if `None`.
    title : str, plot title to use
    ax : axes object, matplotlib axes object to assign figure to.

    Returns
    -------
    ax : matplotlib axes object
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(10, 6))
    else:
        fig = plt.gcf()

    if metric is None:
        metric = sp["outputs"][0]

    if isinstance(metric, str):
        assert metric in sp["outputs"], f"Specified model output '{metric}' not found"

    is_multi_output = len(sp["outputs"]) > 1
    if not index:
        if is_multi_output:
            index = list(sp.analysis[metric].keys())
        else:
            index = list(sp.analysis.keys())

    multi_index = isinstance(index, (list, tuple))
    if multi_index:
        index = [k for k in index if not k.endswith(CONF_COLUMN) and k != "names"]

    if isinstance(index, str):
        index = [index]

    if is_multi_output:
        met_data = sp.analysis[metric]
    else:
        met_data = sp.analysis

    si_shape = met_data[index[0]].shape
    s_data = np.vstack([met_data[k] for k in index if met_data[k].shape == si_shape])

    ax.imshow(s_data)
    fig.colorbar(ax.images[0], ax=ax, shrink=0.9)

    if title is None:
        title = metric

    ax.set_title(title)

    if isinstance(metric, str):
        metric = [metric]

    # Get unique groups (if defined) while maintaining order of group names
    disp_names = sp["groups"]
    if disp_names is None:
        disp_names = sp["names"]

    ax.xaxis.set_ticks(range(0, len(disp_names)))
    ax.xaxis.set_ticklabels(disp_names, rotation=90)

    # Account for indices that may have been filtered out
    # e.g., cannot easily show second-order values at the moment
    disp_idx = s_data.shape[0]
    ax.yaxis.set_ticks(range(0, len(index[0:disp_idx])))
    ax.yaxis.set_ticklabels(index[0:disp_idx])

    return ax



from typing import Dict

import numpy as np

from . import common_args
from . import sobol_sequence
from ..util import scale_samples, read_param_file


def sample(
    problem: Dict,
    N: int,
    delta: float = 0.01,
    seed: int = None,
    skip_values: int = 1024,
) -> np.ndarray:
    """
    Generate matrix of samples for Derivative-based Global Sensitivity Measure (DGSM).

    Start from a QMC (Sobol') sequence and finite difference with delta % steps

    Parameters
    ----------
    problem : dict
        SALib problem specification
    N : int
        Number of samples
    delta : float
        Finite difference step size (percent)
    seed : int or None
        Random seed value
    skip_values : int
        How many values of the Sobol sequence to skip

    Returns
    -------
    np.array : DGSM sequence

    References
    ----------
    1. Sobol', I.M., Kucherenko, S., 2009.
       Derivative based global sensitivity measures and their link with
       global sensitivity indices.
       Mathematics and Computers in Simulation 79, 3009-3017.
       https://doi.org/10.1016/j.matcom.2009.01.023

    2. Sobol', I.M., Kucherenko, S., 2010.
       Derivative based global sensitivity measures.
       Procedia - Social and Behavioral Sciences 2, 7745-7746.
       https://doi.org/10.1016/j.sbspro.2010.05.208
    """
    if seed:
        np.random.seed(seed)

    D = problem["num_vars"]
    bounds = problem["bounds"]

    # Create base sequence - could be any type of sampling
    base_sequence = sobol_sequence.sample(N + skip_values, D)

    # scale before finite differencing
    base_sequence = scale_samples(base_sequence, problem)

    dgsm_sequence = np.empty([N * (D + 1), D])

    index = 0
    for i in range(skip_values, N + skip_values):
        # Copy the initial point
        dgsm_sequence[index, :] = base_sequence[i, :]

        index += 1
        for j in range(D):
            temp = np.zeros(D)
            bnd_j = bounds[j]
            temp[j] = base_sequence[i, j] * delta
            dgsm_sequence[index, :] = base_sequence[i, :] + temp
            dgsm_sequence[index, j] = min(dgsm_sequence[index, j], bnd_j[1])
            dgsm_sequence[index, j] = max(dgsm_sequence[index, j], bnd_j[0])
            index += 1

    return dgsm_sequence


def cli_parse(parser):
    """Add method specific options to CLI parser.

    Parameters
    ----------
    parser : argparse object

    Returns
    -------
    Updated argparse object
    """
    parser.add_argument(
        "-d",
        "--delta",
        type=float,
        required=False,
        default=0.01,
        help="Finite difference step size (percent)",
    )
    return parser


def cli_action(args):
    """Run sampling method

    Parameters
    ----------
    args : argparse namespace
    """
    problem = read_param_file(args.paramfile)
    param_values = sample(problem, args.samples, args.delta, seed=args.seed)
    np.savetxt(
        args.output,
        param_values,
        delimiter=args.delimiter,
        fmt="%." + str(args.precision) + "e",
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)



from scipy.stats import norm

import numpy as np

from . import common_args
from ..util import read_param_file, ResultDict


def analyze(
    problem, X, Y, num_resamples=100, conf_level=0.95, print_to_console=False, seed=None
):
    """Calculates Derivative-based Global Sensitivity Measure on model outputs.

    Returns a dictionary with keys 'vi', 'vi_std', 'dgsm', and 'dgsm_conf',
    where each entry is a list of size D (the number of parameters) containing
    the indices in the same order as the parameter file.


    Notes
    -----
    Compatible with:
        `finite_diff` : :func:`SALib.sample.finite_diff.sample`


    Examples
    --------
        >>> X = finite_diff.sample(problem, 1000)
        >>> Y = Ishigami.evaluate(X)
        >>> Si = dgsm.analyze(problem, Y, print_to_console=False)


    Parameters
    ----------
    problem : dict
        The problem definition
    X : numpy.matrix
        The NumPy matrix containing the model inputs
    Y : numpy.array
        The NumPy array containing the model outputs
    num_resamples : int
        The number of resamples used to compute the confidence
        intervals (default 1000)
    conf_level : float
        The confidence interval level (default 0.95)
    print_to_console : bool
        Print results directly to console (default False)
    seed : int
        Seed to generate a random number


    References
    ----------
    1. Sobol, I. M. and S. Kucherenko (2009). "Derivative based global
           sensitivity measures and their link with global sensitivity
           indices." Mathematics and Computers in Simulation, 79(10):3009-3017,
           doi:10.1016/j.matcom.2009.01.023.
    """
    if seed:
        np.random.seed(seed)

    D = problem["num_vars"]
    Y_size = Y.size

    if Y_size % (D + 1) == 0:
        N = int(Y_size / (D + 1))
    else:
        raise RuntimeError("Incorrect number of samples in model output file.")

    if not 0 < conf_level < 1:
        raise RuntimeError("Confidence level must be between 0-1.")

    dims = (N, D)
    base = np.empty(N)
    X_base = np.empty(dims)
    perturbed = np.empty(dims)
    X_perturbed = np.empty(dims)
    step = D + 1

    base = Y[0:Y_size:step]
    X_base = X[0:Y_size:step, :]

    # First order (+conf.) and Total order (+conf.)
    keys = ("vi", "vi_std", "dgsm", "dgsm_conf")
    S = ResultDict((k, np.empty(D)) for k in keys)
    S["names"] = problem["names"]

    bounds = problem["bounds"]
    for j in range(D):
        perturbed[:, j] = Y[(j + 1) : Y_size : step]
        X_perturbed[:, j] = X[(j + 1) : Y_size : step, j]

        diff = X_perturbed[:, j] - X_base[:, j]
        perturbed_j = perturbed[:, j]
        S["vi"][j], S["vi_std"][j] = calc_vi_stats(base, perturbed_j, diff)
        S["dgsm"][j], S["dgsm_conf"][j] = calc_dgsm(
            base, perturbed_j, diff, bounds[j], num_resamples, conf_level
        )

    if print_to_console:
        print(S.to_df())

    return S


def calc_vi_stats(base, perturbed, x_delta):
    """Calculate v_i mean and std.

    v_i sensitivity measure following Sobol and Kucherenko (2009)
    For comparison, Morris mu* < sqrt(v_i)

    Same as calc_vi_mean but returns standard deviation as well.
    """
    dfdx = ((perturbed - base) / x_delta) ** 2
    return np.mean(dfdx), np.std(dfdx)


def calc_vi_mean(base, perturbed, x_delta):
    """Calculate v_i mean.

    Same as calc_vi_stats but only returns the mean.
    """
    dfdx = ((perturbed - base) / x_delta) ** 2
    return dfdx.mean()


def calc_dgsm(base, perturbed, x_delta, bounds, num_resamples, conf_level):
    """v_i sensitivity measure following Sobol and Kucherenko (2009).
    For comparison, total order S_tot <= dgsm
    """
    D = np.var(base)
    vi = calc_vi_mean(base, perturbed, x_delta)
    dgsm = vi * (bounds[1] - bounds[0]) ** 2 / (D * np.pi**2)

    len_base = len(base)
    s = np.empty(num_resamples)
    r = np.random.randint(len_base, size=(num_resamples, len_base))
    for i in range(num_resamples):
        r_i = r[i]
        s[i] = calc_vi_mean(base[r_i], perturbed[r_i], x_delta[r_i])

    return dgsm, norm.ppf(0.5 + conf_level / 2.0) * s.std(ddof=1)


def cli_parse(parser):
    parser.add_argument(
        "-X",
        "--model-input-file",
        type=str,
        required=True,
        default=None,
        help="Model input file",
    )
    parser.add_argument(
        "-r",
        "--resamples",
        type=int,
        required=False,
        default=1000,
        help="Number of bootstrap resamples for Sobol \
                           confidence intervals",
    )
    return parser


def cli_action(args):
    problem = read_param_file(args.paramfile)

    Y = np.loadtxt(
        args.model_output_file, delimiter=args.delimiter, usecols=(args.column,)
    )
    X = np.loadtxt(args.model_input_file, delimiter=args.delimiter, ndmin=2)
    if len(X.shape) == 1:
        X = X.reshape((len(X), 1))

    analyze(
        problem,
        X,
        Y,
        num_resamples=args.resamples,
        print_to_console=True,
        seed=args.seed,
    )


if __name__ == "__main__":
    common_args.run_cli(cli_parse, cli_action)
